"""eflomal package"""

from collections import Counter
import logging
from operator import itemgetter
from tempfile import NamedTemporaryFile

from .cython import align, read_text, write_text


logger = logging.getLogger(__name__)


class Aligner:
    """Aligner class"""

    def __init__(self, model=3, score_model=0,
                 n_iterations=None, n_samplers=3,
                 rel_iterations=1.0, null_prior=0.2,
                 source_prefix_len=0, source_suffix_len=0,
                 target_prefix_len=0, target_suffix_len=0):
        self.model = model
        self.score_model = score_model
        self.n_iterations = n_iterations
        self.n_samplers = n_samplers
        self.rel_iterations = rel_iterations
        self.null_prior = null_prior
        self.source_prefix_len = source_prefix_len
        self.source_suffix_len = source_suffix_len
        self.target_prefix_len = target_prefix_len
        self.target_suffix_len = target_suffix_len

    def prepare_files(self, src_input_file, src_output_file,
                      trg_input_file, trg_output_file,
                      priors_input_file, priors_output_file):
        """Convert text files to formats used by eflomal

        Inputs should be file objects or any iterables over lines. Outputs
        should be file objects.

        """
        src_index, n_src_sents, src_voc_size = to_eflomal_text_file(
            src_input_file, src_output_file,
            self.source_prefix_len, self.source_suffix_len)
        trg_index, n_trg_sents, trg_voc_size = to_eflomal_text_file(
            trg_input_file, trg_output_file,
            self.target_prefix_len, self.target_suffix_len)
        if n_src_sents != n_trg_sents:
            logger.error(
                'number of sentences differ in input files (%d vs %d)',
                n_src_sents, n_trg_sents)
            raise ValueError('Mismatched file sizes')
        logger.info('Prepared %d sentences for alignment', n_src_sents)
        if priors_input_file:
            logger.info('Reading lexical priors...')
            priors = read_priors(priors_input_file)
            to_eflomal_priors_file(
                priors, src_index, trg_index, priors_output_file)

    def align(self, src_input, trg_input,
              links_filename_fwd=None, links_filename_rev=None,
              scores_filename_fwd=None, scores_filename_rev=None,
              priors_input=None, quiet=True, use_gdb=False):
        """Run alignment for the input"""
        with NamedTemporaryFile('wb') as srcf, \
             NamedTemporaryFile('wb') as trgf, \
             NamedTemporaryFile('w', encoding='utf-8') as priorsf:
            # Write input files for the eflomal binary
            self.prepare_files(
                src_input, srcf, trg_input, trgf, priors_input, priorsf)
            # Run wrapper for the eflomal binary
            align(srcf.name, trgf.name,
                  links_filename_fwd=links_filename_fwd,
                  links_filename_rev=links_filename_rev,
                  statistics_filename=None,
                  scores_filename_fwd=scores_filename_fwd,
                  scores_filename_rev=scores_filename_rev,
                  priors_filename=(None if priors_input is None
                                   else priorsf.name),
                  model=self.model,
                  score_model=self.score_model,
                  n_iterations=self.n_iterations,
                  n_samplers=self.n_samplers,
                  quiet=quiet,
                  rel_iterations=self.rel_iterations,
                  null_prior=self.null_prior,
                  use_gdb=use_gdb)


class TextIndex:
    """Word to index mapping with lowercasing and prefix/suffix removal"""

    def __init__(self, index, prefix_len=0, suffix_len=0):
        self.index = index
        self.prefix_len = prefix_len
        self.suffix_len = suffix_len

    def __len__(self):
        return len(self.index)

    def __getitem__(self, word):
        word = word.lower()
        if self.prefix_len != 0:
            word = word[:self.prefix_len]
        if self.suffix_len != 0:
            word = word[-self.suffix_len:]
        e = self.index.get(word)
        if e is not None:
            e = e + 1
        return e


def to_eflomal_text_file(sentencefile, outfile, prefix_len=0, suffix_len=0):
    """Write sentences to a file read by eflomal binary

    Arguments:

    sentencefile - input text file object
    outfile - output file object
    prefix_len - prefix length to remove
    suffix_len - suffix length to remove

    Returns TextIndex object.

    """
    sents, index = read_text(sentencefile, True, prefix_len, suffix_len)
    n_sents = len(sents)
    voc_size = len(index)
    write_text(outfile, tuple(sents), voc_size)
    return TextIndex(index, prefix_len, suffix_len), n_sents, voc_size


def sentences_from_joint_file(joint_file, index=None):
    """Yield sentences from joint sentences file"""
    for i, line in enumerate(joint_file):
        fields = line.strip().split(' ||| ')
        if len(fields) != 2:
            logger.error('line %d does not contain a single |||'
                         ' separator, or sentence(s) are empty!',
                         i + 1)
            raise ValueError('Invalid joint input line %s' % line)
        if index is None:
            yield fields[0], fields[1]
        else:
            yield fields[index]


def calculate_priors(src_sentences, trg_sentences,
                     fwd_alignments, rev_alignments):
    """Calculate priors from alignments"""
    priors = Counter()
    hmmf_priors = Counter()
    hmmr_priors = Counter()
    ferf_priors = Counter()
    ferr_priors = Counter()
    for lineno, (src_sent, trg_sent, fwd_line, rev_line) in enumerate(
            zip(src_sentences, trg_sentences, fwd_alignments, rev_alignments)):
        src_sent = src_sent.strip().split()
        trg_sent = trg_sent.strip().split()
        fwd_links = [tuple(map(int, s.split('-'))) for s in fwd_line.split()]
        rev_links = [tuple(map(int, s.split('-'))) for s in rev_line.split()]
        for i, j in fwd_links:
            if i >= len(src_sent) or j >= len(trg_sent):
                logger.error('alignment out of bounds in line %d: '
                             '(%d, %d)', lineno + 1, i, j)
                raise ValueError('Invalid input on line %d' % lineno + 1)
            priors[(src_sent[i], trg_sent[j])] += 1

        last_j = -1
        last_i = -1
        for i, j in sorted(fwd_links, key=itemgetter(1)):
            if j != last_j:
                hmmf_priors[i - last_i] += 1
            last_i = i
            last_j = j
        hmmf_priors[len(src_sent) - last_i] += 1

        last_j = -1
        last_i = -1
        for i, j in sorted(rev_links, key=itemgetter(0)):
            if i != last_i:
                hmmr_priors[j - last_j] += 1
            last_i = i
            last_j = j
        hmmr_priors[len(trg_sent) - last_j] += 1

        fwd_fert = Counter(i for i, j in fwd_links)
        rev_fert = Counter(j for i, j in rev_links)
        for i, fert in fwd_fert.items():
            ferf_priors[(src_sent[i], fert)] += 1
        for j, fert in rev_fert.items():
            ferr_priors[(trg_sent[j], fert)] += 1
    # TODO: confirm EOF in all files
    return priors, hmmf_priors, hmmr_priors, ferf_priors, ferr_priors


def write_priors(priorsf, priors_list, hmmf_priors, hmmr_priors,
                 ferf_priors, ferr_priors):
    """Write priors to file object"""
    for (src, trg), alpha in sorted(priors_list.items()):
        print('LEX\t%s\t%s\t%g' % (src, trg, alpha), file=priorsf)
    for (src, fert), alpha in sorted(ferf_priors.items()):
        print('FERF\t%s\t%d\t%g' % (src, fert, alpha), file=priorsf)
    for (trg, fert), alpha in sorted(ferr_priors.items()):
        print('FERR\t%s\t%d\t%g' % (trg, fert, alpha), file=priorsf)
    for jump, alpha in sorted(hmmf_priors.items()):
        print('HMMF\t%d\t%g' % (jump, alpha), file=priorsf)
    for jump, alpha in sorted(hmmr_priors.items()):
        print('HMMR\t%d\t%g' % (jump, alpha), file=priorsf)


def read_priors(priors_file):
    """Load priors from file object"""
    priors_list = []    # list of (srcword, trgword, alpha)
    ferf_priors = []    # list of (wordform, alpha)
    ferr_priors = []    # list of (wordform, alpha)
    hmmf_priors = {}    # dict of jump: alpha
    hmmr_priors = {}    # dict of jump: alpha
    # 5 types of lines valid:
    #
    # LEX   srcword     trgword     alpha   | lexical prior
    # HMMF  jump        alpha               | target-side HMM prior
    # HMMR  jump        alpha               | source-side HMM prior
    # FERF  srcword     fert   alpha        | source-side fertility p.
    # FERR  trgword     fert    alpha       | target-side fertility p.
    for i, line in enumerate(priors_file):
        fields = line.rstrip('\n').split('\t')
        try:
            alpha = float(fields[-1])
        except ValueError as err:
            logger.error('priors line %d contains alpha '
                         'value of "%s" which is not numeric',
                         i+1, fields[2])
            raise err
        if fields[0] == 'LEX' and len(fields) == 4:
            priors_list.append((fields[1], fields[2], alpha))
        elif fields[0] == 'HMMF' and len(fields) == 3:
            hmmf_priors[int(fields[1])] = alpha
        elif fields[0] == 'HMMR' and len(fields) == 3:
            hmmr_priors[int(fields[1])] = alpha
        elif fields[0] == 'FERF' and len(fields) == 4:
            ferf_priors.append((fields[1], int(fields[2]), alpha))
        elif fields[0] == 'FERR' and len(fields) == 4:
            ferr_priors.append((fields[1], int(fields[2]), alpha))
        else:
            logger.error('priors line %d is invalid', i + 1)
            raise ValueError('Invalid input on line %d' % i + 1)
    return priors_list, hmmf_priors, hmmr_priors, ferf_priors, ferr_priors


def to_eflomal_priors_file(priors, src_index, trg_index, outfile):
    """Write priors to a file read by eflomal binary

    Arguments:

    priors - tuple of priors (priors_list, hmmf_priors, hmmr_priors,
             ferf_priors, ferr_priors)
    src_index - vocabulary index for source text
    tgt_index - vocabulary index for target text
    outfile - file object for output

    """
    priors_list, hmmf_priors, hmmr_priors, ferf_priors, ferr_priors = priors
    priors_indexed = {}
    for src_word, trg_word, alpha in priors_list:
        if src_word == '<NULL>':
            e = 0
        else:
            e = src_index[src_word]

        if trg_word == '<NULL>':
            f = 0
        else:
            f = trg_index[trg_word]

        if (e is not None) and (f is not None):
            priors_indexed[(e, f)] = priors_indexed.get((e, f), 0.0) \
                + alpha
    ferf_indexed = {}
    for src_word, fert, alpha in ferf_priors:
        e = src_index[src_word]
        if e is not None:
            ferf_indexed[(e, fert)] = \
                ferf_indexed.get((e, fert), 0.0) + alpha
    ferr_indexed = {}
    for trg_word, fert, alpha in ferr_priors:
        f = trg_index[trg_word]
        if f is not None:
            ferr_indexed[(f, fert)] = \
                ferr_indexed.get((f, fert), 0.0) + alpha
    logger.info('%d (of %d) pairs of lexical priors used',
                len(priors_indexed), len(priors_list))
    print('%d %d %d %d %d %d %d' % (
        len(src_index)+1, len(trg_index)+1, len(priors_indexed),
        len(hmmf_priors), len(hmmr_priors),
        len(ferf_indexed), len(ferr_indexed)),
          file=outfile)
    for (e, f), alpha in sorted(priors_indexed.items()):
        print('%d %d %g' % (e, f, alpha), file=outfile)
    for jump, alpha in sorted(hmmf_priors.items()):
        print('%d %g' % (jump, alpha), file=outfile)
    for jump, alpha in sorted(hmmr_priors.items()):
        print('%d %g' % (jump, alpha), file=outfile)
    for (e, fert), alpha in sorted(ferf_indexed.items()):
        print('%d %d %g' % (e, fert, alpha), file=outfile)
    for (f, fert), alpha in sorted(ferr_indexed.items()):
        print('%d %d %g' % (f, fert, alpha), file=outfile)
    outfile.flush()



cimport cython
from cpython cimport bool
cimport numpy as np
from libc.stdio cimport fprintf, fdopen, fputc, fflush, FILE

import os
import sys
import math
import subprocess
from tempfile import NamedTemporaryFile

import numpy as np


cpdef tuple read_text(pyfile, bool lowercase, int prefix_len, int suffix_len):
    """Read a tokenized text file as a list of indexed sentences.
    
    Optionally transform the vocabulary according to the parameters.
    
    pyfile -- file to read
    lowercase -- if True, all tokens are lowercased
    prefix_len -- if non-zero, all tokens are cut of after so many characters
    suffix_len -- if non-zero, as above, but cutting from the right side

    Returns:
    a tuple (list sents, dict index) containing the actual sentences and the
    string-to-index mapping used.
    """
    cdef:
        np.ndarray[np.uint32_t, ndim=1] sent
        list sents, tokens
        str line, token
        dict index
        int i, n, idx

    index = {}
    sents = []
    for line in pyfile:
        if lowercase:
            tokens = line.lower().split()
        else:
            tokens = line.split()
        n = len(tokens)
        sent = np.empty(n, dtype=np.uint32)

        for i in range(n):
            token = tokens[i]
            if prefix_len != 0: token = token[:prefix_len]
            elif suffix_len != 0: token = token[-suffix_len:]
            idx = index.get(token, -1)
            if idx == -1:
                idx = len(index)
                index[token] = idx
            sent[i] = idx

        sents.append(sent)

    return (sents, index)


cpdef write_text(pyfile, tuple sents, int voc_size):
    """Write a sequence of sentences in the format expected by eflomal

    Arguments:
    pyfile -- Python file object to write to
    sents -- tuple of sentences, each encoded as np.ndarray(uint32)
    voc_size -- size of vocabulary
    """
    cdef int token, i, n
    cdef FILE *f
    cdef np.ndarray[np.uint32_t, ndim=1] sent

    f = fdopen(pyfile.fileno(), 'wb')
    fprintf(f, '%d %d\n', len(sents), voc_size)
    for sent in sents:
        n = len(sent)
        if n < 0x400:
            i = 0
            fprintf(f, '%d', n)
            while i < n:
                fprintf(f, ' %d', sent[i])
                i += 1
            fputc(10, f)
        else:
            fputc(48, f)
            fputc(10, f)
    fflush(f)


def align(
        str source_filename,
        str target_filename,
        str links_filename_fwd=None,
        str links_filename_rev=None,
        str statistics_filename=None,
        str scores_filename_fwd=None,
        str scores_filename_rev=None,
        str priors_filename=None,
        int model=3,
        int score_model=0,
        tuple n_iterations=None,
        int n_samplers=1,
        bool quiet=True,
        double rel_iterations=1.0,
        double null_prior=0.2,
        bool use_gdb=False):
    """Call the eflomal binary to perform word alignment

    Arguments:
    source_filename -- str with source text filename, this and the target
                       text should both be written using write_text()
    target_filename -- str with target text filename
    links_filename_fwd -- if given, write links here (forward direction)
    links_filename_rev -- if given, write links here (reverse direction)
    statistics_filename -- if given, write alignment statistics here
    scores_filename -- if given, write sentence alignment scoeres here
    priors_filename -- if given, read Dirichlet priors from here
    model -- alignment model (1 = IBM1, 2 = HMM, 3 = HMM+fertility)
    n_iterations -- 3-tuple with number of iterations per model, if this is
                    not given the numbers will be computed automatically based
                    on rel_iterations
    n_samplers -- number of independent samplers to run
    quiet -- if True, suppress output
    rel_iterations -- number of iterations relative to the default
    """

    with open(source_filename, 'rb') as f:
        n_sentences = int(f.readline().split()[0])

    if n_iterations is None:
        iters = max(2, int(round(
            rel_iterations*5000 / math.sqrt(n_sentences))))
        iters4 = max(1, iters//4)
        if model == 1:
            n_iterations = (iters, 0, 0)
        elif model == 2:
            n_iterations = (max(2, iters4), iters, 0)
        else:
            n_iterations = (max(2, iters4), iters4, iters)

    executable = os.path.join(os.path.dirname(__file__), 'bin', 'eflomal')
    args = [executable,
            '-m', str(model),
            '-s', source_filename,
            '-t', target_filename,
            '-n', str(n_samplers),
            '-N', str(null_prior),
            '-1', str(n_iterations[0])]
    if quiet: args.append('-q')
    if model >= 2: args.extend(['-2', str(n_iterations[1])])
    if model >= 3: args.extend(['-3', str(n_iterations[2])])
    if links_filename_fwd: args.extend(['-f', links_filename_fwd])
    if links_filename_rev: args.extend(['-r', links_filename_rev])
    if statistics_filename: args.extend(['-S', statistics_filename])
    if score_model > 0: args.extend(['-M', str(score_model)])
    if scores_filename_fwd: args.extend(['-F', scores_filename_fwd])
    if scores_filename_rev: args.extend(['-R', scores_filename_rev])
    if priors_filename: args.extend(['-p', priors_filename])
    if not quiet: sys.stderr.write(' '.join(args) + '\n')
    if use_gdb: args = ['gdb', '-ex=run', '--args'] + args
    subprocess.call(args)




#!/usr/bin/env python3

from setuptools import setup, Extension
from setuptools.command.egg_info import egg_info
from setuptools.command.develop import develop
from setuptools.command.install import install
import subprocess
from Cython.Build import cythonize
import numpy


def compile_and_install_binary():
    subprocess.check_call('make -C src eflomal', shell=True)
    subprocess.check_call('make -C src python-install', shell=True)


class CustomInstall(install):
    """Custom handler for the 'install' command."""
    def run(self):
        compile_and_install_binary()
        super().run()


class CustomEggInfo(egg_info):
    """Custom handler for the 'egg_info' command."""
    def run(self):
        compile_and_install_binary()
        super().run()


class CustomDevelop(develop):
    """Custom handler for the 'develop' command."""
    def run(self):
        compile_and_install_binary()
        super().run()


cyalign_ext=Extension('eflomal.cython', ['python/eflomal/eflomal.pyx'],
                      include_dirs=[numpy.get_include()])

with open('README.md', 'r') as fh:
    long_description = fh.read()

install_requires = ['numpy', 'Cython']
tests_require = ['pytest']

setup(
    name='eflomal',
    version='1.0.0-beta',
    author='Robert Östling',
    url='https://github.com/robertostling/eflomal',
    license='GNU GPLv3',
    description='pip installable eflomal',
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=install_requires,
    tests_require=tests_require,
    extras_require={'test': tests_require},
    packages=['eflomal'],
    package_dir={'': 'python'},
    package_data={
        'eflomal': ['bin/eflomal']
    },
    ext_modules=cythonize(cyalign_ext, language_level='3'),
    scripts=['python/scripts/eflomal-align', 'python/scripts/eflomal-makepriors'],
    cmdclass={'install': CustomInstall, 'develop': CustomDevelop, 'egg_info': CustomEggInfo}
)
